#ifndef types_h
#define types_h

/*
 * USB Device States
 */
enum USBDeviceStates
{
    STATE_DETACHED,
    STATE_ATTACHED,
    STATE_POWERED,
    STATE_DEFAULT,
    STATE_ADDR_PEND,
    STATE_ADDRESSED,
    STATE_CONFIGURED
};

/*
 * SETUP Packet Structure
 */
typedef union SETUP_PKT
{
    // Layout to access data on Control Writes
    unsigned char CtrlDataArray[8];
    
    // Default Layout
    struct
    {
        unsigned char bmRequestType;
        unsigned char bRequest;
        unsigned short wValue;
        unsigned short wIndex;
        unsigned short wLength;
    };

    // GET_DESCRIPTOR Layout
    struct
    {
        unsigned char bmRequestType;
        unsigned char bRequest;
        unsigned char bDscIndex;
        unsigned char bDscType;
        unsigned short wLangID;
        unsigned short wLength;
    };

    // SET_ADDRESS Layout
    struct
    {
        unsigned char bmRequestType;
        unsigned char bRequest;
        unsigned char bDevAddr;
        unsigned char : 8;
        unsigned short wIndex;
        unsigned short wLength;
    };

    // SET_CONFIGURATION Layout
    struct
    {
        unsigned char bmRequestType;
        unsigned char bRequest;
        unsigned char bConfigVal;
        unsigned char : 8;
        unsigned short wIndex;
        unsigned short wLength;
    };
    
    // GET_STATUS Layout
    struct
    {
        unsigned char bmRequestType;
        unsigned char bRequest;
        unsigned short wValue;
        unsigned char bEndpoint;
        unsigned char bReserved;        
        unsigned short wLength;
    };
    
    // SET_FEATURE & CLEAR_FEATURE Layout
    struct
    {
        unsigned char bmRequestType; // @ Offset 0
        unsigned char bRequest;      // @ Offset 1  
        unsigned char bFeatSel;      // @ Offset 2
        unsigned char bReserved;     // @ Offset 3
        unsigned char bEndpoint;     // @ Offset 4
        unsigned char bReserved2;    // @ Offset 5   
        unsigned short wLength;      // @ Offset 6
    };
}SETUP_PKT;

/*
 * Bits of bmRequestType
 */

/* Data Direction */
#define DIR_HOST_DEV    0x00
#define DIR_DEV_HOST    0x80

/* Type */
#define TYPE_STANDARD   0x00
#define TYPE_CLASS      0x20
#define TYPE_VENDOR     0x40
#define TYPE_RESERVED   0x60

#define TYPE_MASK       0x60

/* Recipient */
#define RCP_DEVICE      0x00
#define RCP_INTERFACE   0x01
#define RCP_ENDPOINT    0x02
#define RCP_OTHER       0x03

#define RCP_ENDP_MASK   0x82

/*
 * bRequest Values (Standard Requests)
 */
#define GET_STATUS          0
#define CLEAR_FEATURE       1
#define SET_FEATURE         3
#define SET_ADDRESS         5
#define GET_DESCRIPTOR      6
#define SET_DESCRIPTOR      7
#define GET_CONFIGURATION   8
#define SET_CONFIGURATION   9
#define GET_INTERFACE       10
#define SET_INTERFACE       11
#define SYNCH_FRAME         12

/*
 * Descriptor Types
 */
#define DEVICE 1
#define CONFIGURATION 2
#define STRING 3
#define INTERFACE 4
#define ENDPOINT 5

/*
 * Feature Selectors
 */
#define ENDPOINT_HALT           0
#define DEVICE_REMOTE_WAKEUP    1

/*
 * HID Specific Information
 */
#define HID_INTF_CLASS          3

#define HID_DSC_CLASS           0x21
#define HID_DSC_REPORT          0x22

/* HID Requests */
#define HID_GET_REPORT          0x01
#define HID_GET_IDLE            0x02
#define HID_GET_PROTOCOL        0x03

#define HID_SET_REPORT          0x09
#define HID_SET_IDLE            0x0A
#define HID_SET_PROTOCOL        0x0B

/* Control Transfer Variable Types */
typedef union
{
    ram unsigned char* RamPtr;
    rom unsigned char* RomPtr;
}SOURCE_PTR;

#define REQUEST_OK      0x80
#define REQUEST_ERROR   0x00

#define REQ_CTRL_READ   0x40
#define REQ_CTRL_WRITE  0x00

#define REQ_DATA_RAM    0x20
#define REQ_DATA_ROM    0x00

enum ShortPacketStatus
{
    SHORT_PKT_NONE,
    SHORT_PKT_PEND,
    SHORT_PKT_SENT
};

/*
 *
 * Device Specific Data
 *
 */

/* UCON: USB CONTROL REGISTER */
#define PingPongBuffersReset    UCONbits.PPBRST
#define LiveSingleEndedZero     UCONbits.SE0
#define PacketTransferDisable   UCONbits.PKTDIS
#define USBModuleEnable         UCONbits.USBEN
#define USBResumeEnable         UCONbits.RESUME
#define USBSuspendEnable        UCONbits.SUSPND

/* UCFG: USB CONFIGURATION REGISTER */
#define USB_PULLUP_ENABLE       0x10

#define USB_INT_XCVR            0x00
#define USB_EXT_XCVR            0x08

#define USB_FULL_SPEED          0x04
#define USB_LOW_SPEED           0x00

/* No PingPong for now */


/* USTAT: USB STATUS REGISTER */
typedef union _USTAT
{
    unsigned char AsByte;

    struct
    {
        unsigned char       : 1;
        unsigned char PPBI  : 1;
        unsigned char DIR   : 1;
        unsigned char ENDP  : 4;
        unsigned char       : 1;
    };
}USB_STATUS;

#define USTAT_EP_MASK   0x78

/* UEPn: USB ENDPOINT n CONTROL REGISTER */
#define ENDP_HSHK_ENABLE        0x10
#define ENDP_CTRL_DISABLE       0x08
#define ENDP_OUT_ENABLE         0x04
#define ENDP_IN_ENABLE          0x02
#define ENDP_STALLED            0x01

/* Buffer Descriptor */
typedef struct _BDT_ENTRY
{
    /* BDSTAT */
    union
    {
        unsigned char AsByte;

        /* CPU Mode */
        struct
        {
            unsigned char BCx     : 2;
            unsigned char BSTALL  : 1;
            unsigned char DTSEN   : 1;
            unsigned char INCDIS  : 1;
            unsigned char KEN     : 1;
            unsigned char DTS     : 1;
            unsigned char UOWN    : 1;
        };

        /* SIE Mode */
        struct
        {
            unsigned char BCx     : 2;
            unsigned char PID     : 4;
            unsigned char         : 1;
            unsigned char UOWN    : 1;
        };
    }BDSTAT;

    /* BD Byte Count */
    unsigned char BDCNT;

    /* BD Buffer Address */
    unsigned short BDADDR;
}BDT_ENTRY;

/* BDSTAT Flags */
#define _SIE            0x80
#define _CPU            0x00

#define _DATA0          0x00
#define _DATA1          0x40

#define _DTSEN          0x08
#define _BSTALL         0x04

/* Token Types */
#define USB_TOKEN_SETUP 0x34
#define USB_TOKEN_OUT   0x04
#define USB_TOKEN_IN    0x24

#define USB_PID_MASK    0x3C

/*
 * BD Table Entry Index
 */
#define EP0_OUT         0x00
#define EP0_IN          0x01
#define EP1_OUT         0x02
#define EP1_IN          0x03


/* UIR: USB INTERRUPT STATUS REGISTER */
#define StartOfFrame_Int    UIRbits.SOFIF
#define Stall_Int           UIRbits.STALLIF
#define Idle_Int            UIRbits.IDLEIF
#define Transaction_Int     UIRbits.TRNIF
#define Activity_Int        UIRbits.ACTVIF
#define Error_Int           UIRbits.UERRIF
#define Reset_Int           UIRbits.URSTIF


#endif
