#include <p18cxxx.h>
#include "usb\usb.h"
#include "usb\usbdsc.h"

///////////////
/// USB RAM ///
///////////////

#pragma udata usb_bd_table = 0x0400
BDT_ENTRY BufDscTbl[(1 + USB_ENDP_COUNT)*2];

#pragma udata usb_buf_area = 0x0500
SETUP_PKT SetupPkt;
unsigned char ControlTrfData[8];

/* HID Buffers */
unsigned char hid_in_ep_buf[64];
unsigned char hid_out_ep_buf[64];

///////////////////////
/// State Variables ///
///////////////////////
#pragma udata access usb_state_vars
near enum USBDeviceStates   usb_device_state;           // 0
near unsigned char          usb_active_config;          // 1
near USB_STATUS             usb_transaction_status;     // 2

near SOURCE_PTR             usb_data_source;            // L=3, H=4
near unsigned char          usb_data_length;            // 5
near unsigned char          usb_data_flags;             // 6
near enum ShortPacketStatus usb_short_pkt_status;       // 7

// Auxiliary Variables
near unsigned char          bytes_to_process;
near unsigned char          hid_rpt_rx_len;

#pragma udata
unsigned char*         usb_ram_ptr;

/////////////////
/// Code Area ///
/////////////////

/* Prototypes */
void USBReset(void);
void ServiceSetup(void);
void ServiceOut(void);
void ServiceIn(void);
void CheckForStdRequest(void);
void CheckForHidRequest(void);
void CompleteSetup(void);
void CopyToUSBRAM(void);
void CopyFromUSBRAM(void);
void PrepareForNewSetup(void);
void GetStdDescriptor(void);
void ReportStatus(void);
void StdFeature(void);

void HIDTxReport(unsigned char* buf, unsigned char len);
void HIDRxReport(unsigned char* buf, unsigned char len);

unsigned char IsHIDTxBusy(void);
unsigned char IsHIDRxBusy(void);
/*
 * Name         : USB_SM_Init
 *
 * Synopsis     : void USB_SM_Init(void)
 *
 * Description  :
 *
 */
void USB_SM_Init(void)
{
    usb_device_state = STATE_DETACHED;

    USBReset();

    // Configure Module
    UCFG = USB_PULLUP_ENABLE | USB_INT_XCVR | USB_FULL_SPEED;

    // Enable USB Module and wait for SE0
    USBModuleEnable = 1;

    while (LiveSingleEndedZero == 1)
        ;
}

/*
 * Name         : USB_SM_Run
 *
 * Synopsis     : void USB_SM_Run(void)
 *
 * Description  :
 *
 */
void USB_SM_Run(void)
{
    /* Check for Activity_Int */
    if (Activity_Int == 1)
    {
        /* Exit from Suspend Mode */
        USBSuspendEnable = 0;

        /* Clear Activity_Int flag */
        while (Activity_Int == 1)
            Activity_Int = 0;

        return;
    }

    /* Check for Reset_Int */
    if (Reset_Int == 1)
    {
        USBReset();
        usb_device_state = STATE_DEFAULT;

        /* Initialize Buffer Descriptor Table */
        PrepareForNewSetup();
        return;
    }

    /* Check for Transaction_Int */
    while (Transaction_Int == 1)
    {
        /* Read USTAT */
        usb_transaction_status.AsByte = USTAT;
        
        /* Determine last Buffer Descriptor used */
        //usb_last_bd = (near BDT_ENTRY*)&BufDscTbl[(usb_transaction_status.AsByte >> 2) & 0x1F];

        /* Determine last transaction type */
        if (usb_transaction_status.DIR == 0)
        {
            /* OUT or SETUP */
            /* Read PID */
            if ((BufDscTbl[EP0_OUT].BDSTAT.AsByte & USB_PID_MASK) == USB_TOKEN_SETUP)
            {                
                ServiceSetup();
                PacketTransferDisable = 0;
            }
            else
            {
                ServiceOut();
            }
        }
        else
        {
            /* IN */
            ServiceIn();
        }
        
        /* Clear Transaction_Int to advance USTAT FIFO */
        Transaction_Int = 0;
    }

    /* Check for Idle State */
    if (Idle_Int == 1)
    {
        Idle_Int = 0;
        USBSuspendEnable = 1;
    }
    else
    {
        /* Ignore all other interrupts */
        UEIR = 0x00;
        UIR &= 0x0F;
    }
}

void USBReset(void)
{
    usb_active_config = 0x00;
    usb_data_flags = REQUEST_ERROR;
    usb_short_pkt_status = SHORT_PKT_NONE;

    // Clear USTAT FIFO
    while (Transaction_Int == 1)
        Transaction_Int = 0;

    // Clear USB Interrupts
    UEIR = 0x00;
    UIR = 0x00;
    PIR2bits.USBIF = 0;

    // Reset all ping-pong buffers
    PingPongBuffersReset = 1;

    // Clear Buffer Descriptor Table
    _asm
        lfsr 0, 0x0400
NEXT:   clrf POSTINC0, 0
        btfss FSR0L, 4, 0
        bra NEXT
    _endasm

    PingPongBuffersReset = 0;

    // Configure Endpoints
    _asm
        lfsr 0, 0x0F71
N2:     clrf POSTINC0, 0
        btfss FSR0L, 7, 0
        bra N2
    _endasm

    UEP0 = ENDP_HSHK_ENABLE | ENDP_OUT_ENABLE | ENDP_IN_ENABLE;

    // Enable Packet Reception
    PacketTransferDisable = 0;
}

void ServiceSetup(void)
{
    if (BufDscTbl[EP0_IN].BDSTAT.UOWN == 1)
        BufDscTbl[EP0_IN].BDSTAT.AsByte = _CPU;

    usb_data_flags = REQUEST_ERROR;
    usb_short_pkt_status = SHORT_PKT_NONE;
    
    /* Process Request */
    if ((SetupPkt.bmRequestType & TYPE_MASK) == TYPE_STANDARD)
        CheckForStdRequest();
    else
        CheckForHidRequest();

    CompleteSetup();
}

void ServiceOut(void)
{
    if (!(usb_transaction_status.AsByte & USTAT_EP_MASK))
    {
        /* Check transfer status variables to determine current transfer's type */
        if (usb_data_flags & REQ_CTRL_READ)
        {
            /* Control Read Status Stage */
            PrepareForNewSetup();
        }
        else
        {
            /* Control Write Data Stage */

            // Copy data to user's buffer
            CopyFromUSBRAM();

            // Check if there is more data to receive
            if (usb_data_length == 0)
            {
                // No accept more data, STALL
                BufDscTbl[EP0_OUT].BDSTAT.AsByte = _SIE | _BSTALL;
                BufDscTbl[EP0_OUT].BDCNT = 0;
            }
            else
            {
                BufDscTbl[EP0_OUT].BDCNT = sizeof(ControlTrfData);

                /* Toggle DTS */
                BufDscTbl[EP0_OUT].BDSTAT.AsByte &= _DATA1;
                BufDscTbl[EP0_OUT].BDSTAT.AsByte ^= _DATA1;
                BufDscTbl[EP0_OUT].BDSTAT.AsByte |= _SIE | _DTSEN;
            }
        }
    }
}

void ServiceIn(void)
{
    if (!(usb_transaction_status.AsByte & USTAT_EP_MASK))
    {
        /* Check transfer status variables to determine current transfer's type */
        if (usb_data_flags & REQ_CTRL_READ)
        {
            /* Control Read Data Stage */
            CopyToUSBRAM();

            /* Check short packet status */
            if (usb_short_pkt_status == SHORT_PKT_SENT)
            {
                /* The host is querying more data than pacted, STALL */
                BufDscTbl[EP0_IN].BDSTAT.AsByte = _SIE | _BSTALL;
            }
            else
            {
                /* Toggle DTS */
                BufDscTbl[EP0_IN].BDSTAT.AsByte &= _DATA1;
                BufDscTbl[EP0_IN].BDSTAT.AsByte ^= _DATA1;
                BufDscTbl[EP0_IN].BDSTAT.AsByte |= _SIE | _DTSEN;
            }
        }
        else
        {
            /* Control Write Status Stage */

            // Check for SET_ADDRESS pending processing
            if (usb_device_state == STATE_ADDR_PEND)
            {
                UADDR = SetupPkt.bDevAddr;

                if (UADDR == 0)
                    usb_device_state = STATE_DEFAULT;
                else
                    usb_device_state = STATE_ADDRESSED;
            }

            PrepareForNewSetup();
        }
    }
}

void CheckForStdRequest(void)
{
    switch (SetupPkt.bRequest)
    {
        case GET_DESCRIPTOR:
        {
            GetStdDescriptor();
        }
        break;

        case SET_ADDRESS:
        {
            usb_device_state = STATE_ADDR_PEND;
            usb_data_flags = REQUEST_OK | REQ_CTRL_WRITE;
        }
        break;

        case SET_CONFIGURATION:
        {
            if (SetupPkt.bConfigVal == 1)
            {
                /* Configure Endpoint 1 */
                UEP1 = ENDP_HSHK_ENABLE | ENDP_IN_ENABLE | ENDP_OUT_ENABLE | ENDP_CTRL_DISABLE;

                BufDscTbl[EP1_OUT].BDADDR = (unsigned short)&hid_out_ep_buf[0];
                BufDscTbl[EP1_OUT].BDCNT = sizeof(hid_out_ep_buf);
                BufDscTbl[EP1_OUT].BDSTAT.AsByte = _SIE | _DTSEN | _DATA0;

                BufDscTbl[EP1_IN].BDADDR = (unsigned short)&hid_in_ep_buf[0];
                BufDscTbl[EP1_IN].BDCNT = sizeof(hid_in_ep_buf);
                BufDscTbl[EP1_IN].BDSTAT.AsByte = _CPU | _DTSEN | _DATA1;

                usb_device_state = STATE_CONFIGURED;
                usb_data_flags = REQUEST_OK | REQ_CTRL_WRITE;
            }
            else if (SetupPkt.bConfigVal == 0)
            {
                UEP1 = 0x00;

                usb_device_state = STATE_ADDRESSED;
                usb_data_flags = REQUEST_OK | REQ_CTRL_WRITE;
            }
        }
        break;

        case GET_STATUS:
        {
            ReportStatus();
        }
        break;

        case CLEAR_FEATURE:
        case SET_FEATURE:
        {
            StdFeature();
        }
        break;

        case GET_CONFIGURATION:
        {
            /* Copy Current config to USB RAM */
            usb_data_source.RamPtr = (ram unsigned char*)&usb_active_config;
            usb_data_length = 1;
            usb_data_flags = REQUEST_OK | REQ_CTRL_READ | REQ_DATA_RAM;
        }
        break;

        case GET_INTERFACE:
        {
            ControlTrfData[0] = 0;  // Initialize content

            usb_data_source.RamPtr = (ram unsigned char*)&ControlTrfData[0];
            usb_data_length = 1;
            usb_data_flags = REQUEST_OK | REQ_CTRL_READ | REQ_DATA_RAM;
        }
        break;

        case SET_INTERFACE:
        {
            usb_data_flags = REQUEST_OK | REQ_CTRL_WRITE;
        }
        break;
    }
}

void CheckForHidRequest(void)
{
    switch (SetupPkt.bRequest)
    {
        case HID_SET_IDLE:
        {
            usb_data_flags = REQUEST_OK | REQ_CTRL_WRITE;
        }
        break;

        case HID_GET_REPORT:
        {
            usb_data_source.RamPtr = (ram unsigned char*)&hid_in_ep_buf[0];
            usb_data_length = sizeof(hid_in_ep_buf);
            usb_data_flags = REQUEST_OK | REQ_CTRL_READ | REQ_DATA_RAM;
        }
        break;
    }
}

void CompleteSetup(void)
{
    /* Check if request was accepted */
    if (!(usb_data_flags & REQUEST_OK))
    {
        /* Request Error */

        /* STALL handshake on IN endpoint */
        BufDscTbl[EP0_IN].BDSTAT.AsByte = _SIE | _BSTALL;

        /* Only accept SETUP on OUT endpoint */
        BufDscTbl[EP0_OUT].BDADDR = (unsigned short)&SetupPkt;
        BufDscTbl[EP0_OUT].BDCNT = sizeof(SetupPkt);
        BufDscTbl[EP0_OUT].BDSTAT.AsByte = _SIE | _DTSEN | _DATA0 | _BSTALL;
    }
    else
    {
        /* Check for Control Transfer Type */
        if (usb_data_flags & REQ_CTRL_READ)
        {
            /* Control Read Transfer */

            // Correct usb_data_length
            if (SetupPkt.wLength < usb_data_length)
                usb_data_length = SetupPkt.wLength;

            // Copy data to USB RAM from usb_data_source
            CopyToUSBRAM();

            // Arm EP0_IN as Data1 to send data to host
            BufDscTbl[EP0_IN].BDADDR = (unsigned short)&ControlTrfData[0];
            BufDscTbl[EP0_IN].BDSTAT.AsByte = _SIE | _DTSEN | _DATA1;

            // Arm EP0_OUT to receive the Status Stage or a new SETUP
            BufDscTbl[EP0_OUT].BDADDR = (unsigned short)&SetupPkt;
            BufDscTbl[EP0_OUT].BDCNT = sizeof(SetupPkt);
            BufDscTbl[EP0_OUT].BDSTAT.AsByte = _SIE | _DTSEN | _DATA1;
        }
        else
        {
            /* Control Write Transfer */

            // Arm EP0_OUT to receive data from host (to SetupPkt)
            BufDscTbl[EP0_OUT].BDADDR = (unsigned short)&SetupPkt;
            BufDscTbl[EP0_OUT].BDCNT = sizeof(SetupPkt);
            BufDscTbl[EP0_OUT].BDSTAT.AsByte = _SIE | _DTSEN | _DATA1;

            // Arm EP0_IN to send the Status Stage to the host
            BufDscTbl[EP0_IN].BDCNT = 0;
            BufDscTbl[EP0_IN].BDSTAT.AsByte = _SIE | _DTSEN | _DATA1;
        }
    }
}

void CopyToUSBRAM(void)
{
    if (usb_data_length > sizeof(ControlTrfData))
        bytes_to_process = sizeof(ControlTrfData);
    else
    {
        bytes_to_process = usb_data_length;

        if (usb_short_pkt_status < SHORT_PKT_SENT)
            usb_short_pkt_status++;
    }

    usb_data_length -= bytes_to_process;

    BufDscTbl[EP0_IN].BDCNT = bytes_to_process;

    usb_ram_ptr = (unsigned char*)&ControlTrfData[0];

    if (usb_data_flags & REQ_DATA_RAM)
    {
        while (bytes_to_process)
        {
            *usb_ram_ptr++ = *usb_data_source.RamPtr++;
            bytes_to_process--;
        }
    }
    else
    {
        while (bytes_to_process)
        {
            *usb_ram_ptr++ = *usb_data_source.RomPtr++;
            bytes_to_process--;
        }
    }
}

void CopyFromUSBRAM(void)
{
    /* Copy received data from USB RAM to usb_data_source.RamPtr */
    bytes_to_process = BufDscTbl[EP0_OUT].BDCNT;

    /* Update usb_data_length */
    usb_data_length -= bytes_to_process;
    usb_ram_ptr = (unsigned char*)&SetupPkt.CtrlDataArray[0];

    while (bytes_to_process)
    {
        *usb_data_source.RamPtr++ = *usb_ram_ptr++;
        bytes_to_process--;
    }
}

void PrepareForNewSetup(void)
{
    // Prepare to receive a new SETUP
    BufDscTbl[EP0_OUT].BDADDR = (unsigned short)&SetupPkt;
    BufDscTbl[EP0_OUT].BDCNT = sizeof(SetupPkt);
    BufDscTbl[EP0_OUT].BDSTAT.AsByte = _SIE | _DTSEN | _DATA0 | _BSTALL;

    BufDscTbl[EP0_IN].BDSTAT.AsByte = _CPU;
}

void GetStdDescriptor(void)
{
    switch (SetupPkt.bDscType)
    {
        case DEVICE:
        {
            /* Copy Device Descriptor to USB RAM */
            usb_data_source.RomPtr = (rom unsigned char*)&DeviceDescriptor[0];
            usb_data_length = (unsigned char)*usb_data_source.RomPtr;
            usb_data_flags = REQUEST_OK | REQ_CTRL_READ | REQ_DATA_ROM;
        }
        break;

        case CONFIGURATION:
        {
            if (SetupPkt.bDscIndex == 0)
            {
                /* Copy Configuration Descriptor to USB RAM */
                usb_data_source.RomPtr = (rom unsigned char*)&ConfigDescriptor[0];
                usb_data_length = 41;
                usb_data_flags = REQUEST_OK | REQ_CTRL_READ | REQ_DATA_ROM;
            }
        }
        break;

        case STRING:
        {
            if (SetupPkt.bDscIndex < 4)
            {
                /* Copy requested String Descriptor to USB RAM */
                usb_data_source.RomPtr = (rom unsigned char*)StringList[SetupPkt.bDscIndex];
                usb_data_length = (unsigned char)*usb_data_source.RomPtr;
                usb_data_flags = REQUEST_OK | REQ_CTRL_READ | REQ_DATA_ROM;
            }
        }
        break;

        case HID_DSC_CLASS:
        {
            /* Copy HID Descriptor to USB RAM */
            usb_data_source.RomPtr = (rom unsigned char*)&ConfigDescriptor[18];
            usb_data_length = (unsigned char)*usb_data_source.RomPtr;
            usb_data_flags = REQUEST_OK | REQ_CTRL_READ | REQ_DATA_ROM;
        }
        break;

        case HID_DSC_REPORT:
        {
            /* Copy HID Report Descriptor to USB RAM */
            usb_data_source.RomPtr = (rom unsigned char*)&HidReportDsc[0];
            usb_data_length = 29;
            usb_data_flags = REQUEST_OK | REQ_CTRL_READ | REQ_DATA_ROM;
        }
        break;
    }
}

void ReportStatus(void)
{
    _asm
        /* Initialize ControlTrfData[0] & ControlTrfData[1] to zero */
        movlb 5     // Select Bank 05 (ControlTrfData is @ 0x0508
        clrf 8, 1   // RAM[0x0508] = 0x00
        clrf 9, 1   // RAM[0x0509] = 0x00

        /* Determine if request recipient is Endpoint (bmRequestType == 0x82 */
        movlw 0x82
        xorwf 0, 0, 1 // xorwf reg @ 0x0500 with 0x82 -> W
        bnz nstall    // If W != 0, bmRequestType != 0x82

        /* Endpoint Recipient, point to BDSTAT */
        lfsr 0, 0x0400

        /* Determine the endpoint of interest */
        btfsc 4, 0, 1 // Test bit 0 of bEndpoint and skip if clear
        bsf FSR0L, 3, 0 // If endpoint is EP1 then set FSR0 |= 0x08

        /* Determine the endpoint direction */
        btfsc 4, 7, 1 // Test bit 7 of bEndpoint to detect if is IN
        bsf FSR0L, 2, 0 // If endpoint is IN then set FSR0 |= 0x04

        /* Read BDSTAT */
        movf INDF0, 0, 0 // Copy RAM[FSR0] -> W

        /* Determine if _BSTALL & W is not zero (Buffer Stalled) */
        andlw 0x04
        bz nstall

        /* If Stalled then update ControlTrfData[0] |= 0x01 */
        bsf 8, 0, 1 // Set bit 0 of register at 0x0508 [Banked]

        /* Update usb status variables */
        /*
            usb_data_source.RamPtr = (ram unsigned char*)&ControlTrfData[0];
            usb_data_length = 2;
            usb_data_flags = REQUEST_OK | REQ_CTRL_READ | REQ_DATA_RAM;
        */

        // The next 4 inst are for
        // usb_data_source.RamPtr = (ram unsigned char*)&ControlTrfData[0];
nstall: movlw 0x08
        movwf 3, 0  // usb_data_source.RamPtr.LowPart = 0x08
        movlw 0x05
        movwf 4, 0  // usb_data_source.RamPtr.HighPart = 0x05

        // usb_data_length = 2;
        movlw 2
        movwf 5, 0  // usb_data_length = 2

        // usb_data_flags = REQUEST_OK | REQ_CTRL_READ | REQ_DATA_RAM;
        movlw 0xE0
        movwf 6, 0 // usb_data_flags = ...

    _endasm
}

void StdFeature(void)
{
    /* Set/Clear Feature for Endpoint 1
     *  0x0400 + 8 EP1_OUT:BDSTAT   0___0001    =>  1000
     *  0x0400 + C EP1_IN:BDSTAT    1___0001    =>  1100
     */
    _asm
        // Check if are met all conditions if (...)

        /* Point to Bank 5 (0x0500) where is located SetupPkt */
        movlb 5

        /* Determine if SetupPkt.bmRequestType == 2 */
        movlw 2
        xorwf 0, 0, 1 // xorwf reg @ 0x0500 with 2 -> W
        bnz exit // if W != 0 then exit

        /* Determine if SetupPkt.bEndpoint == 1 */
        movlw 1
        andwf 4, 0, 1 // andwf reg @ 0x0504 with 1 -> W
        bz exit // if W != 1 then exit

        /* Determine if SetupPkt.bFeatSel == ENDPOINT_HALT (0) */
        movf 2, 0, 1 // copy reg @ 0x0502 to W and update flags
        bnz exit // if W != 0 then exit

        movlw 0x80 // REQUEST_OK | REQ_CTRL_WRITE
        movwf 6, 0 // usb_data_flags @ pos 6 in access bank

        /* Point to BDSTAT byte of BufDscTbl[EP1_OUT] using FSR0 */
        lfsr 0, 0x0408

        /* Determine endpoint's direction from SetupPkt.bEndpoint (bit 7) */
        btfsc 4, 7, 1 // Test bit 7 of reg @ 0x0504 and skip if clear
        bsf FSR0L, 2, 0 // Set bit 2 of FSR0L using Access Bank if bit 7 is set (changes to EP1_IN)

        /* Determine if Standard request was SET_FEATURE */
        btfsc 1, 1, 1 // Test bit 1 of reg @ 0x0501 and skip if clear
        bra b_set

        // If not jump then request was CLEAR_FEATURE
        /* Determine endpoint's direction from SetupPkt.bEndpoint (bit 7) */
        btfss 4, 7, 1 // Test bit 7 of reg @ 0x0504 and skip if set
        bra e1_out  // If EP1_OUT then arm as _SIE | _DTSEN | _DATA0;
        clrf INDF0, 0   // Else set BDSTAT1_IN as _CPU
        bra exit

e1_out: movlw  0x88 // _SIE | _DTSEN | _DATA0;
        movwf INDF0, 0
        bra exit

b_set:  movlw 0x84 // _SIE | _BSTALL // SET_FEATURE -> Write _SIE | _BSTALL to INDF0 (BDSTAT of EP1)
        movwf INDF0, 0
exit:
    _endasm
}

void HIDTxReport(unsigned char* buf, unsigned char len)
{
    BufDscTbl[EP1_IN].BDCNT = len;

    usb_ram_ptr = (unsigned char*)&hid_in_ep_buf[0];

    while (len)
    {
        *usb_ram_ptr++ = *buf++;
        len--;
    }

    BufDscTbl[EP1_IN].BDSTAT.AsByte &= _DATA1;
    BufDscTbl[EP1_IN].BDSTAT.AsByte ^= _DATA1;
    BufDscTbl[EP1_IN].BDSTAT.AsByte |= _SIE | _DTSEN;
}

void HIDRxReport(unsigned char* buf, unsigned char len)
{
    if(!IsHIDRxBusy())
    {
        /*
         * Adjust the expected number of bytes to equal
         * the actual number of bytes received.
         */
        hid_rpt_rx_len = BufDscTbl[EP1_OUT].BDCNT;

        if (len < hid_rpt_rx_len)
            hid_rpt_rx_len = len;

        /*
         * Copy data from dual-ram buffer to user's buffer
         */

        usb_ram_ptr = (unsigned char*)&hid_out_ep_buf[0];

        while (len)
        {
            *buf++ = *usb_ram_ptr++;
            len--;
        }

        /*
         * Prepare dual-ram buffer for next OUT transaction
         */
        BufDscTbl[EP1_OUT].BDCNT = sizeof(hid_out_ep_buf);

        BufDscTbl[EP1_OUT].BDSTAT.AsByte &= _DATA1;
        BufDscTbl[EP1_OUT].BDSTAT.AsByte ^= _DATA1;
        BufDscTbl[EP1_OUT].BDSTAT.AsByte |= _SIE | _DTSEN;
    }
}

unsigned char IsHIDTxBusy(void)
{
    return BufDscTbl[EP1_IN].BDSTAT.UOWN;
}

unsigned char IsHIDRxBusy(void)
{
    return BufDscTbl[EP1_OUT].BDSTAT.UOWN;
}

unsigned char GetRxCount(void)
{
    return hid_rpt_rx_len;
}